#pragma once

#include "parser.hpp"
#include <cassert>
#include <unordered_map>

class Generator {
public:
    inline explicit Generator(NodeProg prog)
        : m_prog(std::move(prog))
    {
    }

    void gen_term(const NodeTerm* term)
    {
        struct TermVisitor {
            Generator* gen;
            void operator()(const NodeTermIntLit* term_int_lit) const
            {
                if(!term_int_lit->int_lit.value.has_value()) {
                    std::cerr << "Optional value not set in NodeTermIntLit." << std::endl;
                    exit(EXIT_FAILURE);
                }
                gen->m_output << "    LDM ";
                gen->m_output <<term_int_lit->int_lit.value.value() << "\n";
            }
            void operator()(const NodeTermIdent* term_ident) const
            {
                auto it = std::find_if(gen->m_vars.cbegin(), gen->m_vars.cend(), [&](const Var& var){
                    return var.name == term_ident->ident.value.value();
                });
                if (it == gen->m_vars.cend()) {
                    std::cerr << "Undeclared identifier: " << term_ident->ident.value.value() << std::endl;
                    exit(EXIT_FAILURE);
                }
                gen->point((*it).page_loc, (*it).reg_loc, (*it).word_loc);

                gen->m_output << "    RDM\n";
                gen->m_output << "; Read " << term_ident->ident.value.value() <<" from memory\n";
            }
            void operator()(const NodeTermParen* term_paren) const
            {
                gen->gen_expr(term_paren->expr);
            }
        };
        TermVisitor visitor({ .gen = this });
        std::visit(visitor, term->var);
    }

    void gen_bin_expr(const NodeBinExpr* bin_expr)
    {
        struct BinExprVisitor {
            Generator* gen;
            void operator()(const NodeBinExprSub* sub) const
            {
                // Handle the left-hand side
                bool lhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(sub->lhs->var)) {
                    NodeBinExpr* lhs_expr = std::get<NodeBinExpr*>(sub->lhs->var);
                    gen->gen_bin_expr(lhs_expr);
                    lhs_processed = true;
                    std::cerr << "Processed nested binary expression for lhs" << std::endl;
                    gen->m_output << "    CLC\n";  // Store the result of lhs
                } else {

                }
                // Handle the right-hand side
                bool rhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(sub->rhs->var)) {
                    NodeBinExpr* rhs_expr = std::get<NodeBinExpr*>(sub->rhs->var);
                    gen->gen_bin_expr(rhs_expr);
                    rhs_processed = true;
                    std::cerr << "Processed nested binary expression for rhs" << std::endl;

                } else {
                    gen->gen_expr(sub->rhs);
                }
                if (!lhs_processed) {
                    gen->m_output << "    XCH 5\n";  // Store the result of lhs
                    gen->gen_expr(sub->lhs);
                }


                // Now, if either lhs or rhs were already processed or neither of them are nested expressions, add them
                if (lhs_processed || rhs_processed || (!lhs_processed && !rhs_processed)) {
                    gen->m_output << "    SUB 5\n";
                    if (!rhs_processed) {
                        gen->m_output << "    XCH 5\n";
                    }
                    return;
                }
            }
            void operator()(const NodeBinExprAdd* add) const {
                // Handle the left-hand side
                bool lhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(add->lhs->var)) {
                    NodeBinExpr* lhs_expr = std::get<NodeBinExpr*>(add->lhs->var);
                    gen->gen_bin_expr(lhs_expr);
                    lhs_processed = true;
                    std::cerr << "Processed nested binary expression for lhs" << std::endl;
                    gen->m_output << "    CLC\n";  // Store the result of lhs
                } else {

                }
                // Handle the right-hand side
                bool rhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(add->rhs->var)) {
                    NodeBinExpr* rhs_expr = std::get<NodeBinExpr*>(add->rhs->var);
                    gen->gen_bin_expr(rhs_expr);
                    rhs_processed = true;
                    std::cerr << "Processed nested binary expression for rhs" << std::endl;


                } else {
                    gen->gen_expr(add->rhs);
                }
                if (!lhs_processed) {
                    gen->m_output << "    XCH 5\n";  // Store the result of lhs
                    gen->gen_expr(add->lhs);
                }

                // Now, if either lhs or rhs were already processed or neither of them are nested expressions, add them
                if (lhs_processed || rhs_processed || (!lhs_processed && !rhs_processed)) {
                    gen->m_output << "    ADD 5\n";
                    gen->m_output << "    XCH 5\n";
                    return;
                }
            }
            void operator()(const NodeBinExprMulti* multi) const
            {
                // Handle the left-hand side
                bool lhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(multi->lhs->var)) {
                    NodeBinExpr* lhs_expr = std::get<NodeBinExpr*>(multi->lhs->var);
                    gen->gen_bin_expr(lhs_expr);
                    lhs_processed = true;
                    std::cerr << "Processed nested binary expression for lhs" << std::endl;
                } else {
                    gen->gen_expr(multi->lhs);
                }

                gen->m_output << "    XCH 3\n";  // Store the result of lhs

                // Handle the right-hand side
                bool rhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(multi->rhs->var)) {
                    NodeBinExpr* rhs_expr = std::get<NodeBinExpr*>(multi->rhs->var);
                    gen->gen_bin_expr(rhs_expr);
                    rhs_processed = true;
                    std::cerr << "Processed nested binary expression for rhs" << std::endl;
                } else {
                    gen->gen_expr(multi->rhs);
                }

                gen->m_output << "    XCH 4\n";  // Store the result of rhs

                // If either the lhs or rhs was a nested binary expression or neither were, we go directly to multiplication.
                if (lhs_processed || rhs_processed || (!lhs_processed && !rhs_processed)) {
                    gen->m_output << "    JMS MULT\n";
                    return;
                }
            }
            void operator()(const NodeBinExprDiv* div) const
            {
                // Handle the left-hand side
                bool lhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(div->lhs->var)) {
                    NodeBinExpr* lhs_expr = std::get<NodeBinExpr*>(div->lhs->var);
                    gen->gen_bin_expr(lhs_expr);
                    lhs_processed = true;
                    std::cerr << "Processed nested binary expression for lhs" << std::endl;
                } else {
                    gen->gen_expr(div->lhs);
                }

                gen->m_output << "    XCH 4\n";  // Store the result of lhs

                // Handle the right-hand side
                bool rhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(div->rhs->var)) {
                    NodeBinExpr* rhs_expr = std::get<NodeBinExpr*>(div->rhs->var);
                    gen->gen_bin_expr(rhs_expr);
                    rhs_processed = true;
                    std::cerr << "Processed nested binary expression for rhs" << std::endl;
                } else {
                    gen->gen_expr(div->rhs);
                }

                gen->m_output << "    XCH 5\n";  // Store the result of rhs

                // If either the lhs or rhs was a nested binary expression or neither were, we go directly to division.
                if (lhs_processed || rhs_processed || (!lhs_processed && !rhs_processed)) {
                    gen->m_output << "    JMS DIV\n";
                    gen->m_output << "    XCH 0\n";
                    return;
                }
            }
            void operator()(const NodeBinExprPwr* pwr) const
            {
                // Handle the base (lhs)
                bool lhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(pwr->lhs->var)) {
                    NodeBinExpr* lhs_expr = std::get<NodeBinExpr*>(pwr->lhs->var);
                    gen->gen_bin_expr(lhs_expr);
                    lhs_processed = true;
                } else {
                    gen->gen_expr(pwr->lhs);
                }

                gen->m_output << "    XCH 4\n";  // Store the result of lhs (base)

                // Handle the exponent (rhs)
                bool rhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(pwr->rhs->var)) {
                    NodeBinExpr* rhs_expr = std::get<NodeBinExpr*>(pwr->rhs->var);
                    gen->gen_bin_expr(rhs_expr);
                    rhs_processed = true;
                } else {
                    gen->gen_expr(pwr->rhs);
                }

                gen->m_output << "    XCH 6\n";  // Store the result of rhs (exponent)

                // If either the lhs or rhs was a nested binary expression or neither were, we go directly to power.
                if (lhs_processed || rhs_processed || (!lhs_processed && !rhs_processed)) {
                    gen->m_output << "    JMS PWR\n";
                    return;
                }
            }
            void operator()(const NodeBinExprCmpr* cmpr) const {
                // Handle the lhs
                bool lhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(cmpr->lhs->var)) {
                    NodeBinExpr* lhs_expr = std::get<NodeBinExpr*>(cmpr->lhs->var);
                    gen->gen_bin_expr(lhs_expr);
                    lhs_processed = true;
                } else {
                    gen->gen_expr(cmpr->lhs);
                }

                gen->m_output << "    XCH 5\n";  // Store the result of lhs

                // Handle the rhs
                bool rhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(cmpr->rhs->var)) {
                    NodeBinExpr* rhs_expr = std::get<NodeBinExpr*>(cmpr->rhs->var);
                    gen->gen_bin_expr(rhs_expr);
                    rhs_processed = true;
                } else {
                    gen->gen_expr(cmpr->rhs);
                }

                gen->m_output << "    SUB 5\n";  // Perform the subtraction operation

                int currentIfCount = gen->m_ifCounter;
                gen->m_output << "    JCN 1100, ";
                // If either the lhs or rhs was a nested binary expression or neither were, return as they are compared directly.
                if (lhs_processed || rhs_processed || (!lhs_processed && !rhs_processed)) {
                    return;
                }


                // ... [rest of your code for special cases]

                // Like previously, consider breaking out special cases into helper functions for readability and maintainability.
            }
            void operator()(const NodeBinExprGT* gt) const {

                // Handle the rhs
                bool rhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(gt->rhs->var)) {
                    NodeBinExpr* rhs_expr = std::get<NodeBinExpr*>(gt->rhs->var);
                    gen->gen_bin_expr(rhs_expr);
                    rhs_processed = true;
                } else {
                    gen->gen_expr(gt->rhs);
                }

                gen->m_output << "    XCH 5\n";  // Store the result of lhs

                // Handle the lhs
                bool lhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(gt->lhs->var)) {
                    NodeBinExpr* lhs_expr = std::get<NodeBinExpr*>(gt->lhs->var);
                    gen->gen_bin_expr(lhs_expr);
                    lhs_processed = true;
                } else {
                    gen->gen_expr(gt->lhs);
                }



                gen->m_output << "    SUB 5\n";  // Perform the subtraction operation

                int currentIfCount = gen->m_ifCounter;

                // Jump if CN
                gen->m_output << "    JCN 0010, ";

                // If either the lhs or rhs was a nested binary expression or neither were, return as they are compared directly.
                if (lhs_processed || rhs_processed || (!lhs_processed && !rhs_processed)) {
                    return;
                }

            }
            void operator()(const NodeBinExprGTEQ* gteq) const {

                // Handle the rhs
                bool rhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(gteq->rhs->var)) {
                    NodeBinExpr* rhs_expr = std::get<NodeBinExpr*>(gteq->rhs->var);
                    gen->gen_bin_expr(rhs_expr);
                    rhs_processed = true;
                } else {
                    gen->gen_expr(gteq->rhs);
                }

                gen->m_output << "    XCH 5\n";  // Store the result of lhs

                // Handle the lhs
                bool lhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(gteq->lhs->var)) {
                    NodeBinExpr* lhs_expr = std::get<NodeBinExpr*>(gteq->lhs->var);
                    gen->gen_bin_expr(lhs_expr);
                    lhs_processed = true;
                } else {
                    gen->gen_expr(gteq->lhs);
                }

                gen->m_output << "    SUB 5\n";  // Perform the subtraction operation

                int currentIfCount = gen->m_ifCounter;
                gen->m_output << "    JCN CN||AZ, "; // Jump if CN || Jump if AZ\n";

                // If either the lhs or rhs was a nested binary expression or neither were, return as they are compared directly.
                if (lhs_processed || rhs_processed || (!lhs_processed && !rhs_processed)) {
                    return;
                }
            }
            void operator()(const NodeBinExprLT* lt) const {


                // Handle the rhs
                bool rhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(lt->rhs->var)) {
                    NodeBinExpr* rhs_expr = std::get<NodeBinExpr*>(lt->rhs->var);
                    gen->gen_bin_expr(rhs_expr);
                    rhs_processed = true;
                } else {
                    gen->gen_expr(lt->rhs);
                }

                gen->m_output << "    XCH 5\n";  // Store the result of lhs

                // Handle the lhs
                bool lhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(lt->lhs->var)) {
                    NodeBinExpr* lhs_expr = std::get<NodeBinExpr*>(lt->lhs->var);
                    gen->gen_bin_expr(lhs_expr);
                    lhs_processed = true;
                } else {
                    gen->gen_expr(lt->lhs);
                }

                gen->m_output << "    SUB 5\n";  // Perform the subtraction operation

                int currentWhileCount = gen->m_whileCounter;

                // jump if CZ
                gen->m_output << "    JCN 1010, ";

                // If either the lhs or rhs was a nested binary expression or neither were, return as they are compared directly.
                if (lhs_processed || rhs_processed || (!lhs_processed && !rhs_processed)) {
                    return;
                }
            }
            void operator()(const NodeBinExprLTEQ* lteq) const {

                // Handle the rhs
                bool rhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(lteq->rhs->var)) {
                    NodeBinExpr* rhs_expr = std::get<NodeBinExpr*>(lteq->rhs->var);
                    gen->gen_bin_expr(rhs_expr);
                    rhs_processed = true;
                } else {
                    gen->gen_expr(lteq->rhs);
                }

                gen->m_output << "    XCH 5\n";  // Store the result of lhs

                // Handle the lhs
                bool lhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(lteq->lhs->var)) {
                    NodeBinExpr* lhs_expr = std::get<NodeBinExpr*>(lteq->lhs->var);
                    gen->gen_bin_expr(lhs_expr);
                    lhs_processed = true;
                } else {
                    gen->gen_expr(lteq->lhs);
                }

                gen->m_output << "    SUB 5\n";  // Perform the subtraction operation

                int currentIfCount = gen->m_ifCounter;
                gen->m_output << "    JCN CZ||AZ, ";

                // If either the lhs or rhs was a nested binary expression or neither were, return as they are compared directly.
                if (lhs_processed || rhs_processed || (!lhs_processed && !rhs_processed)) {
                    return;
                }

                // ... [rest of your code for special cases]

                // Like previously, consider breaking out special cases into helper functions for readability and maintainability.
            }
            void operator()(const NodeBinExprCmprNEQ* cmpr) const {
                // Handle the lhs
                bool lhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(cmpr->lhs->var)) {
                    NodeBinExpr* lhs_expr = std::get<NodeBinExpr*>(cmpr->lhs->var);
                    gen->gen_bin_expr(lhs_expr);
                    lhs_processed = true;
                } else {
                    gen->gen_expr(cmpr->lhs);
                }

                gen->m_output << "    XCH 5\n";  // Store the result of lhs

                // Handle the rhs
                bool rhs_processed = false;
                if (std::holds_alternative<NodeBinExpr*>(cmpr->rhs->var)) {
                    NodeBinExpr* rhs_expr = std::get<NodeBinExpr*>(cmpr->rhs->var);
                    gen->gen_bin_expr(rhs_expr);
                    rhs_processed = true;
                } else {
                    gen->gen_expr(cmpr->rhs);
                }

                gen->m_output << "    SUB 3\n";  // Perform the subtraction operation

                int currentIfCount = gen->m_ifCounter;

                //Jump AZ
                gen->m_output << "    JCN 0100, ";
                // If either the lhs or rhs was a nested binary expression or neither were, return as they are compared directly.
                if (lhs_processed || rhs_processed || (!lhs_processed && !rhs_processed)) {
                    return;
                }


                // ... [rest of your code for special cases]

                // Like previously, consider breaking out special cases into helper functions for readability and maintainability.
            }
        };

        BinExprVisitor visitor { .gen = this };
        std::visit(visitor, bin_expr->var);
    }

    void gen_expr(const NodeExpr* expr)
    {
        struct ExprVisitor {
            Generator* gen;
            void operator()(const NodeTerm* term) const
            {
                gen->gen_term(term);
            }
            void operator()(const NodeBinExpr* bin_expr) const
            {
                gen->gen_bin_expr(bin_expr);
                gen->m_output << "    LD 5\n";

            }
        };

        ExprVisitor visitor { .gen = this };
        std::visit(visitor, expr->var);
    }

    void gen_stmt(const NodeStmt* stmt)
    {
        struct StmtVisitor {
            Generator* gen;
            void operator()(const NodeStmtExit* stmt_exit) const
            {
                gen->gen_expr(stmt_exit->expr);
                gen->m_output << "    NOP\n";
            }
            void operator()(const NodeStmtAsgn* stmt_asgn) const
            {
                auto it = std::find_if(gen->m_vars.cbegin(), gen->m_vars.cend(), [&](const Var& var){
                    return var.name == stmt_asgn->ident.value.value();
                });

                if (it == gen->m_vars.cend()) {
                    // The variable wasn't found. Raise an error.
                    std::cerr << "Error: Variable '" << stmt_asgn->ident.value.value() << "' not declared." << std::endl;
                    exit(EXIT_FAILURE);  // Exit the program with an error code. Adjust this based on your desired error handling mechanism.
                }

                gen->gen_expr(stmt_asgn->expr);

                gen->point((*it).page_loc, (*it).reg_loc, (*it).word_loc);

                gen->m_output << "    WRM\n";
            }
            void operator()(const NodeStmtIf* stmt_if) const
            {
                int currentIfCount = ++(gen->m_ifCounter);
                // Generate the condition
                gen->gen_expr(stmt_if->expr);
                // If the condition is false, jump to the else part
                gen->m_output << "Else" << currentIfCount << ":\n"; //

                gen->gen_scope(stmt_if->scope);

                gen->m_ifCounter++;
                // If there's an else part, jump to the end of the if-else construct after executing the 'if' body
                gen->m_output << "Else" << currentIfCount << ":\n"; // else start label
                if (stmt_if->_else) {

                    gen->gen_scope(stmt_if->_else->scope);
                }
                gen->m_output << "_EndIf" << currentIfCount << ":\n"; // else start label
            }
            void operator()(const NodeStmtWhile* stmt_while) const
            {
                int currentWhileCount = ++(gen->m_whileCounter);

                gen->m_output << "While" << currentWhileCount << ":\n"; //
                gen->gen_expr(stmt_while->expr);

                gen->m_output << "EndWhile" << currentWhileCount << ":\n"; //


                gen->gen_scope(stmt_while->scope);

                gen->m_output << "    JCN 0, While" << currentWhileCount << ":\n"; //

                gen->m_whileCounter++;

                gen->m_output << "EndWhile" << currentWhileCount << ":\n";
            }
            void operator()(const NodeStmtLet* stmt_let) const
            {
                auto it = std::find_if(gen->m_vars.cbegin(), gen->m_vars.cend(), [&](const Var& var){
                    return var.name == stmt_let->ident.value.value();
                });

                if (it != gen->m_vars.cend()) {
                    std::cerr << "Identifier already used: " << stmt_let->ident.value.value() << std::endl;
                    exit(EXIT_FAILURE);
                }
                gen->m_vars.push_back({.name = stmt_let->ident.value.value(), .word_loc = gen->m_word_loc, .reg_loc = gen->m_reg_loc, .page_loc = gen->m_page_loc});

                it = gen->m_vars.end() - 1;

                gen->point((*it).page_loc, (*it).reg_loc, (*it).word_loc);

                gen->gen_expr(stmt_let->expr);

                gen->incrementMemoryLocation();
                gen->m_output << "    WRM\n";

                gen->m_output << "; Wrote to " << stmt_let->ident.value.value() << "\n";
            }
            void operator()(const NodeStmtInc* stmt_inc) const
            {
                auto it = std::find_if(gen->m_vars.cbegin(), gen->m_vars.cend(), [&](const Var& var){
                    return var.name == stmt_inc->ident.value.value();
                });
                // Handle the left-hand side
                gen->point((*it).page_loc, (*it).reg_loc, (*it).word_loc);
                gen->m_output << "    RDM\n";
                gen->m_output << "    IAC\n";

                gen->point((*it).page_loc, (*it).reg_loc, (*it).word_loc);
                gen->m_output << "    WRM\n";
            }
            void operator()(const NodeStmtDec* stmt_dec) const
            {
                auto it = std::find_if(gen->m_vars.cbegin(), gen->m_vars.cend(), [&](const Var& var){
                    return var.name == stmt_dec->ident.value.value();
                });
                // Handle the left-hand side
                auto& var = (*it);

                gen->point(var.page_loc, var.reg_loc, var.word_loc);
                gen->m_output << "    RDM\n";
                gen->m_output << "    DAC\n";

                gen->point(var.page_loc, var.reg_loc, var.word_loc);
                gen->m_output << "    WRM\n";
            }
            void operator()(const NodeStmtAddAsgn* stmt_addasgn) const
            {
                auto it = std::find_if(gen->m_vars.cbegin(), gen->m_vars.cend(), [&](const Var& var){
                    return var.name == stmt_addasgn->ident.value.value();
                });
                // Handle the left-hand side
                auto& var = (*it);
                gen->gen_expr(stmt_addasgn->expr);
                gen->m_output << "    XCH 5\n";

                gen->point(var.page_loc, var.reg_loc, var.word_loc);
                gen->m_output << "    ADD 5\n";
                gen->m_output << "    WRM\n";
                gen->m_output << "; Add and Assigned\n";
            }
            void operator()(const NodeStmtSubAsgn* stmt_subasgn) const
            {
                auto it = std::find_if(gen->m_vars.cbegin(), gen->m_vars.cend(), [&](const Var& var){
                    return var.name == stmt_subasgn->ident.value.value();
                });
                // Handle the left-hand side
                auto& var = (*it);
                gen->gen_expr(stmt_subasgn->expr);

                gen->point(var.page_loc, var.reg_loc, var.word_loc);
                gen->m_output << "    WRM\n";
                gen->m_output << "; Assigned\n";
            }
            void operator()(const NodeStmtScope* stmt_scope) const
            {
                gen->gen_scope(stmt_scope);
            }
            void operator()(const NodeStmtDecFunc* stmt_decfunc) const
            {
                int currentFuncCount = ++(gen->m_funcCounter);

                auto it = std::find_if(gen->m_funcs.cbegin(), gen->m_funcs.cend(), [&](const Funcs& func){
                    return func.name == stmt_decfunc->ident.value.value();
                });

                if (it != gen->m_funcs.cend()) {
                    std::cerr << "Function already declared: " << stmt_decfunc->ident.value.value() << std::endl;
                    exit(EXIT_FAILURE);
                }

                gen->m_funcs.push_back({.name = stmt_decfunc->ident.value.value(), .number = currentFuncCount});

                gen->m_output << "FUNC" << currentFuncCount << ":" << std::endl;

                gen->gen_scope(stmt_decfunc->scope);

                gen->m_output << "    BBL 0\n";

            }
            void operator()(const NodeStmtCallFunc* stmt_callfunc) const {
                auto it = std::find_if(gen->m_funcs.cbegin(), gen->m_funcs.cend(), [&](const Funcs& var){
                    return var.name == stmt_callfunc->ident.value.value();
                });

                if (it == gen->m_funcs.cend()) {
                    // The variable wasn't found. Raise an error.
                    std::cerr << "Error: Variable '" << stmt_callfunc->ident.value.value() << "' not declared." << std::endl;
                    exit(EXIT_FAILURE);  // Exit the program with an error code. Adjust this based on your desired error handling mechanism.
                }
                int funcCount = (*it).number;

                gen->m_output << "    JMS FUNC" << funcCount << std::endl;
            }
        };

        StmtVisitor visitor { .gen = this };
        std::visit(visitor, stmt->var);
    }

    void gen_scope(const NodeStmtScope* scope)
    {
        begin_scope();
        for (const NodeStmt* stmt : scope->stmts) {
            gen_stmt(stmt);
        }
        end_scope();
    }

    [[nodiscard]] std::string gen_prog()
    {
        m_output << "STRT \n";

        for (const NodeStmt* stmt : m_prog.stmts) {
            gen_stmt(stmt);
        }
        m_output << "    JUN END\n";

        //Power Subroutine
        m_output << "PWR ";
        m_output << "LDM 1\n";
        m_output << "    XCH 3\n";
        m_output << "    XCH 6\n";
        m_output << "    JCN 3, PEND\n";
        m_output << "PWR2 ";
        m_output << "JMS MULT\n";
        m_output << "    LD 1\n";
        m_output << "    XCH 3\n";
        m_output << "    LD 6\n";
        m_output << "    DAC\n";
        m_output << "    JCN 3, PEND\n";
        m_output << "    XCH 6\n";
        m_output << "    JUN PWR2\n";
        m_output << "PEND ";
        m_output << "LD 3\n";
        m_output << "    XCH 0\n";
        m_output << "    BBL 1\n";

        //Divide Subroutine
        m_output << "DIV ";
        m_output << "LD  3\n";
        m_output << "    XCH 0\n";
        m_output << "DIV2 ";
        m_output << "XCH 4\n";
        m_output << "    SUB 3\n";
        m_output << "    JCN 10, DEND\n";
        m_output << "    XCH 4\n";
        m_output << "    XCH 0\n";
        m_output << "    DAC\n";
        m_output << "    XCH 0\n";
        m_output << "    JUN DIV2\n";
        m_output << "DEND ";
        m_output << "BBL 1\n";

        //Multiply Subroutine
        m_output << "MULT ";
        m_output << "LDM 0\n";
        m_output << "    XCH 5\n";
        m_output << "    LD 3\n";
        m_output << "    SUB 4\n";
        m_output << "    JCN 2, GRTR\n";
        m_output << "    CLC\n";
        m_output << "    LD 3\n";
        m_output << "    XCH 0\n";
        m_output << "    LD 4\n";
        m_output << "    JUN MULT2\n";
        m_output << "GRTR ";
        m_output << "CLC\n";
        m_output << "    LDM 1\n";
        m_output << "    XCH 6\n";
        m_output << "    LD 4\n";
        m_output << "    XCH 0\n";
        m_output << "    LD 3\n";
        m_output << "MULT2 ";
        m_output << "ADD 5\n";
        m_output << "    XCH 5\n";
        m_output << "    LD 0\n";
        m_output << "    DAC 0\n";
        m_output << "    CLC\n";
        m_output << "    JCN 4, MEND\n";
        m_output << "    XCH 0\n";
        m_output << "    LD 6\n";
        m_output << "    JCN 4, LT\n";
        m_output << "    LD 3\n";
        m_output << "    JUN MULT2\n";
        m_output << "LT XCH 4\n";
        m_output << "    JUN MULT2\n";
        m_output << "MEND ";
        m_output << "BBL 1\n";


        m_output << "END NOP\n";
        return m_output.str();
    }

private:

    void incrementMemoryLocation() {
        if (m_word_loc == 15) {
            m_word_loc = 0;
            if (m_reg_loc == 3) {
                m_reg_loc = 0;
                std::cerr << "Next Ram Bank" << std::endl;
                std::cerr << "Add LDM " << m_page_loc << std::endl;
                std::cerr << "And DCL " << std::endl;
                if (m_page_loc == 16) {
                    std::cerr << "Out of Memory" << std::endl;
                    exit(EXIT_FAILURE);
                } else {
                    m_page_loc++;
                }
            } else {
                m_reg_loc++;
            }
        } else {
            m_word_loc++;
        }
    }
    void decrementMemoryLocation() {
        if (m_word_loc == 0) {
            m_word_loc = 15;
            if (m_reg_loc == 0) {
                m_reg_loc = 3;
                if (m_page_loc == 0) {
                    std::cerr << "Stack Underflow" << std::endl;
                    exit(EXIT_FAILURE);
                } else {
                    m_page_loc--;
                }
            } else {
                m_reg_loc--;
            }
        } else {
            m_word_loc--;
        }
    }

    void begin_scope()
    {
        m_scopes.push_back(m_vars.size());
    }

    void end_scope(){
        size_t pop_count = m_vars.size() - m_scopes.back();
        //std::cerr << pop_count << std::endl;
        for (int i = 0; i < pop_count; i++) {
            m_vars.pop_back();
            decrementMemoryLocation();
        }
        m_scopes.pop_back();

    }

    void point(unsigned long page, unsigned long reg, unsigned long word) {
        std::size_t combined_value = (reg * 16) + word;
        //m_output << "    LDM " << page << "\n";
        //m_output << "    DCL\n";
        m_output << "    FIM 0 " << combined_value << "\n";
        m_output << "    SRC 0\n";
    }

    struct Var {
        std::string name;
        size_t word_loc;
        size_t reg_loc;
        size_t page_loc;
    };

    struct Funcs {
        std::string name;
        int number;
    };

    int m_ifCounter = 0;
    int m_whileCounter = 0;
    int m_funcCounter = 0;

    const NodeProg m_prog;

    std::stringstream m_output;
    size_t m_word_loc = 0;
    size_t m_reg_loc = 0;
    size_t m_page_loc = 0;
    std::vector<Var> m_vars {};
    std::vector<Funcs> m_funcs {};
    std::vector<size_t> m_scopes {};
};

